package pkg

import (
	"context"
	"html/template"
	"net/http"
)

type ContactEntry struct {
	Name  string
	Email string
	Phone string
}

type ContactsIndexPage struct {
	Contacts []*ContactEntry
}

func ContactsHandler() http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		conn := getConn(r)
		company := getCompany(r)
		if r.Method == "POST" {
			locale := getLocale(r)
			form := newContactForm(r.Context(), conn, locale)
			if err := form.Parse(r); err != nil {
				http.Error(w, err.Error(), http.StatusBadRequest)
				return
			}
			if err := verifyCsrfTokenValid(r); err != nil {
				http.Error(w, err.Error(), http.StatusForbidden)
				return
			}
			if form.Validate(r.Context(), conn) {
				conn.MustExec(r.Context(), "insert into contact (company_id, business_name, vatin, trade_name, phone, email, web, address, province, city, postal_code, country_code) values ($1, $2, ($12 || $3)::vatin, $4, parse_packed_phone_number($5, $12), $6, $7, $8, $9, $10, $11, $12)", company.Id, form.BusinessName, form.VATIN, form.TradeName, form.Phone, form.Email, form.Web, form.Address, form.City, form.Province, form.PostalCode, form.Country)
				http.Redirect(w, r, "/company/"+company.Slug+"/contacts", http.StatusSeeOther)
			} else {
				mustRenderAppTemplate(w, r, "contacts-new.gohtml", form)
			}
		} else {
			page := &ContactsIndexPage{
				Contacts: mustGetContactEntries(r.Context(), conn, company),
			}
			mustRenderAppTemplate(w, r, "contacts-index.gohtml", page)
		}
	})
}

func mustGetContactEntries(ctx context.Context, conn *Conn, company *Company) []*ContactEntry {
	rows, err := conn.Query(ctx, "select business_name, email, phone from contact where company_id = $1 order by business_name", company.Id)
	if err != nil {
		panic(err)
	}
	defer rows.Close()

	var entries []*ContactEntry
	for rows.Next() {
		entry := &ContactEntry{}
		err = rows.Scan(&entry.Name, &entry.Email, &entry.Phone)
		if err != nil {
			panic(err)
		}
		entries = append(entries, entry)
	}
	if rows.Err() != nil {
		panic(rows.Err())
	}

	return entries
}

type contactForm struct {
	locale       *Locale
	BusinessName *InputField
	VATIN        *InputField
	TradeName    *InputField
	Phone        *InputField
	Email        *InputField
	Web          *InputField
	Address      *InputField
	City         *InputField
	Province     *InputField
	PostalCode   *InputField
	Country      *SelectField
}

func newContactForm(ctx context.Context, conn *Conn, locale *Locale) *contactForm {
	return &contactForm{
		locale: locale,
		BusinessName: &InputField{
			Name:     "business_name",
			Label:    pgettext("input", "Business name", locale),
			Type:     "text",
			Required: true,
			Attributes: []template.HTMLAttr{
				`autocomplete="organization"`,
			},
		},
		VATIN: &InputField{
			Name:     "vatin",
			Label:    pgettext("input", "VAT number", locale),
			Type:     "text",
			Required: true,
		},
		TradeName: &InputField{
			Name:  "trade_name",
			Label: pgettext("input", "Trade name", locale),
			Type:  "text",
		},
		Phone: &InputField{
			Name:     "phone",
			Label:    pgettext("input", "Phone", locale),
			Type:     "tel",
			Required: true,
			Attributes: []template.HTMLAttr{
				`autocomplete="tel"`,
			},
		},
		Email: &InputField{
			Name:     "email",
			Label:    pgettext("input", "Email", locale),
			Type:     "email",
			Required: true,
			Attributes: []template.HTMLAttr{
				`autocomplete="email"`,
			},
		},
		Web: &InputField{
			Name:  "web",
			Label: pgettext("input", "Web", locale),
			Type:  "url",
			Attributes: []template.HTMLAttr{
				`autocomplete="url"`,
			},
		},
		Address: &InputField{
			Name:     "address",
			Label:    pgettext("input", "Address", locale),
			Type:     "text",
			Required: true,
			Attributes: []template.HTMLAttr{
				`autocomplete="address-line1"`,
			},
		},
		City: &InputField{
			Name:     "city",
			Label:    pgettext("input", "City", locale),
			Type:     "text",
			Required: true,
		},
		Province: &InputField{
			Name:     "province",
			Label:    pgettext("input", "Province", locale),
			Type:     "text",
			Required: true,
		},
		PostalCode: &InputField{
			Name:     "postal_code",
			Label:    pgettext("input", "Postal code", locale),
			Type:     "text",
			Required: true,
			Attributes: []template.HTMLAttr{
				`autocomplete="postal-code"`,
			},
		},
		Country: &SelectField{
			Name:     "country",
			Label:    pgettext("input", "Country", locale),
			Options:  mustGetCountryOptions(ctx, conn, locale),
			Selected: "ES",
			Attributes: []template.HTMLAttr{
				`autocomplete="country"`,
			},
		},
	}
}

func (form *contactForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	form.BusinessName.FillValue(r)
	form.VATIN.FillValue(r)
	form.TradeName.FillValue(r)
	form.Phone.FillValue(r)
	form.Email.FillValue(r)
	form.Web.FillValue(r)
	form.Address.FillValue(r)
	form.City.FillValue(r)
	form.Province.FillValue(r)
	form.PostalCode.FillValue(r)
	form.Country.FillValue(r)
	return nil
}

func (form *contactForm) Validate(ctx context.Context, conn *Conn) bool {
	validator := newFormValidator()
	validator.CheckRequiredInput(form.BusinessName, gettext("Business name can not be empty.", form.locale))
	if validator.CheckRequiredInput(form.VATIN, gettext("VAT number can not be empty.", form.locale)) {
		validator.CheckValidVATINInput(form.VATIN, form.Country.Selected, gettext("This value is not a valid VAT number.", form.locale))
	}
	if validator.CheckRequiredInput(form.Phone, gettext("Phone can not be empty.", form.locale)) {
		validator.CheckValidPhoneInput(form.Phone, form.Country.Selected, gettext("This value is not a valid phone number.", form.locale))
	}
	if validator.CheckRequiredInput(form.Email, gettext("Email can not be empty.", form.locale)) {
		validator.CheckValidEmailInput(form.Email, gettext("This value is not a valid email. It should be like name@domain.com.", form.locale))
	}
	if form.Web.Val != "" {
		validator.checkValidURL(form.Web, gettext("This value is not a valid web address. It should be like https://domain.com/.", form.locale))
	}
	validator.CheckRequiredInput(form.Address, gettext("Address can not be empty.", form.locale))
	validator.CheckRequiredInput(form.City, gettext("City can not be empty.", form.locale))
	validator.CheckRequiredInput(form.Province, gettext("Province can not be empty.", form.locale))
	if validator.CheckRequiredInput(form.PostalCode, gettext("Postal code can not be empty.", form.locale)) {
		validator.CheckValidPostalCode(ctx, conn, form.PostalCode, form.Country.Selected, gettext("This value is not a valid postal code.", form.locale))
	}
	validator.CheckValidSelectOption(form.Country, gettext("Selected country is not valid.", form.locale))
	return validator.AllOK()
}

func NewContactHandler() http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		locale := getLocale(r)
		conn := getConn(r)
		form := newContactForm(r.Context(), conn, locale)
		mustRenderAppTemplate(w, r, "contacts-new.gohtml", form)
	})
}
